"""
NumpyArrayItem.

This module defines the ``NumpyArrayItem`` class used to serialize instances of
``numpy.Array``, using binary protocols.
"""

from __future__ import annotations

from functools import cached_property
from io import BytesIO
from typing import TYPE_CHECKING

from .item import (
    Item,
    ItemTypeError,
    b64_str_to_bytes,
    bytes_to_b64_str,
    lazy_is_instance,
)

if TYPE_CHECKING:
    import numpy


class NumpyArrayItem(Item):
    """Serialize instances of ``numpy.Array``, using binary protocols."""

    def __init__(self, array_b64_str: str):
        """
        Initialize a ``NumpyArrayItem``.

        Parameters
        ----------
        array_b64_str : str
            The raw bytes of the array in the ``numpy`` serialization format, encoded in
            a base64 string.
        """
        self.array_b64_str = array_b64_str

    @cached_property
    def __raw__(self) -> numpy.ndarray:
        """Get the value from the ``NumpyArrayItem``."""
        import numpy

        array_bytes = b64_str_to_bytes(self.array_b64_str)

        with BytesIO(array_bytes) as stream:
            return numpy.load(stream)

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``NumpyArrayItem`` instance."""
        return {
            "representation": {
                "media_type": "application/json",
                "value": self.__raw__.tolist(),
            }
        }

    @classmethod
    def factory(cls, value: numpy.ndarray, /) -> NumpyArrayItem:
        """
        Create a new ``NumpyArrayItem`` from an instance of ``numpy.Array``.

        It uses binary protocols.

        Parameters
        ----------
        value: ``numpy.Array``
            The value to serialize.

        Returns
        -------
        NumpyArrayItem
            A new ``NumpyArrayItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` is not an instance of ``numpy.Array``.
        """
        if not lazy_is_instance(value, "numpy.ndarray"):
            raise ItemTypeError(f"Type '{value.__class__}' is not supported.")

        import numpy

        with BytesIO() as stream:
            numpy.save(stream, value, allow_pickle=False)

            array_bytes = stream.getvalue()
            array_b64_str = bytes_to_b64_str(array_bytes)

        instance = cls(array_b64_str)
        instance.__raw__ = value

        return instance



"""
JSONableItem.

This module defines the ``JSONableItem`` class used to serialize objects using the
``JSON`` format.
"""

from __future__ import annotations

from json import dumps, loads
from typing import Any

from .item import Item, ItemTypeError


class JSONableItem(Item):
    """Serialize objects using the ``JSON`` format."""

    def __init__(self, value: Any):
        """
        Initialize a ``JSONableItem``.

        Parameters
        ----------
        value : Any
            The value.
        """
        self.value = value

    @property
    def __raw__(self):
        """Get the value from the ``JSONableItem`` instance."""
        return self.value

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``JSONableItem`` instance."""
        return {
            "representation": {
                "media_type": "application/json",
                "value": self.value,
            }
        }

    @classmethod
    def factory(cls, value: Any, /) -> JSONableItem:
        """
        Create a new ``JSONableItem`` from ``value`` using the ``JSON`` format.

        Parameters
        ----------
        value: Any
            The value to serialize.

        Returns
        -------
        JSONableItem
            A new ``JSONableItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` cannot be serialized using the ``JSON`` format.
        """
        try:
            value = loads(dumps(value))
        except TypeError:
            raise ItemTypeError(f"Type '{value.__class__}' is not supported.") from None

        return cls(value)



"""
MatplotlibFigureItem.

This module defines the ``MatplotlibFigureItem`` class used to serialize instances of
``matplotlib.Figure``, using binary protocols.
"""

from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

from .item import ItemTypeError, bytes_to_b64_str, lazy_is_instance, switch_mpl_backend
from .pickle_item import PickleItem

if TYPE_CHECKING:
    from matplotlib.figure import Figure


class MatplotlibFigureItem(PickleItem):
    """Serialize instances of ``matplotlib.Figure``, using binary protocols."""

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``MatplotlibFigureItem`` instance."""
        with switch_mpl_backend(), BytesIO() as stream:
            self.__raw__.savefig(stream, format="svg", bbox_inches="tight")

            figure_bytes = stream.getvalue()
            figure_b64_str = bytes_to_b64_str(figure_bytes)

        return {
            "representation": {
                "media_type": "image/svg+xml;base64",
                "value": figure_b64_str,
            }
        }

    @classmethod
    def factory(cls, value: Figure, /) -> MatplotlibFigureItem:
        """
        Create a new ``MatplotlibFigureItem`` from an instance of ``matplotlib.Figure``.

        It uses binary protocols.

        Parameters
        ----------
        value : ``matplotlib.Figure``
            The value to serialize.

        Returns
        -------
        MatplotlibFigureItem
            A new ``MatplotlibFigureItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` is not an instance of ``matplotlib.Figure``.
        """
        if not lazy_is_instance(value, "matplotlib.figure.Figure"):
            raise ItemTypeError(f"Type '{value.__class__}' is not supported.")

        return super().factory(value)



"""
SklearnBaseEstimatorItem.

This module defines the ``SklearnBaseEstimatorItem`` class used to serialize instances
of ``sklearn.base.BaseEstimator``.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .item import (
    Item,
    ItemTypeError,
    b64_str_to_bytes,
    bytes_to_b64_str,
    lazy_is_instance,
)

if TYPE_CHECKING:
    import sklearn.base


class SklearnBaseEstimatorItem(Item):
    """Serialize instances of ``sklearn.base.BaseEstimator``."""

    def __init__(
        self,
        estimator_html_repr: str,
        estimator_skops_b64_str: str,
        estimator_skops_untrusted_types: list[str],
    ):
        """
        Initialize a ``SklearnBaseEstimatorItem``.

        Parameters
        ----------
        estimator_html_repr : str
            The HTML representation of the ``sklearn`` estimator.
        estimator_skops_b64_str : str
            The ``skops`` serialization of the ``sklearn`` estimator, encoded in a
            base64 str.
        estimator_skops_untrusted_types: list[str]
            The list of untrusted types in the ``skops`` serialization.
        """
        self.estimator_html_repr = estimator_html_repr
        self.estimator_skops_b64_str = estimator_skops_b64_str
        self.estimator_skops_untrusted_types = estimator_skops_untrusted_types

    @property
    def __raw__(self) -> sklearn.base.BaseEstimator:
        """Get the value from the ``SklearnBaseEstimatorItem`` instance."""
        import skops.io

        estimator_skops_bytes = b64_str_to_bytes(self.estimator_skops_b64_str)

        return skops.io.loads(
            data=estimator_skops_bytes,
            trusted=self.estimator_skops_untrusted_types,
        )

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``SklearnBaseEstimatorItem`` instance."""
        return {
            "representation": {
                "media_type": "application/vnd.sklearn.estimator+html",
                "value": self.estimator_html_repr,
            }
        }

    @classmethod
    def factory(cls, value: sklearn.base.BaseEstimator, /) -> SklearnBaseEstimatorItem:
        """
        Create a new ``SklearnBaseEstimatorItem``.

        Create a new ``SklearnBaseEstimatorItem`` from an instance of
        ``sklearn.base.BaseEstimator``.

        Parameters
        ----------
        value : ``sklearn.base.BaseEstimator``
            The value to serialize.

        Returns
        -------
        SklearnBaseEstimatorItem
            A new ``SklearnBaseEstimatorItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` is not an instance of ``sklearn.base.BaseEstimator``.
        """
        if not lazy_is_instance(value, "sklearn.base.BaseEstimator"):
            raise ItemTypeError(f"Type '{value.__class__}' is not supported.")

        import sklearn.utils
        import skops.io

        estimator_html_repr = sklearn.utils.estimator_html_repr(value)
        estimator_skops_bytes = skops.io.dumps(value)
        estimator_skops_b64_str = bytes_to_b64_str(estimator_skops_bytes)
        estimator_skops_untrusted_types = skops.io.get_untrusted_types(
            data=estimator_skops_bytes
        )

        return cls(
            estimator_html_repr,
            estimator_skops_b64_str,
            estimator_skops_untrusted_types,
        )



"""
MediaItem.

This module defines the ``MediaType`` class used to serialize string media.
"""

from __future__ import annotations

from enum import Enum
from typing import Any

from .item import Item, ItemTypeError


class MediaType(Enum):  # noqa: D101
    HTML = "text/html"
    MARKDOWN = "text/markdown"
    SVG = "image/svg+xml"


class MediaItem(Item):
    """Serialize string media."""

    def __init__(self, media: str, media_type: str):
        """
        Initialize a ``MediaItem``.

        Parameters
        ----------
        media : str
            The media to serialize.
        media_type : str
            The MIME type of the media content.
        """
        self.media = media
        self.media_type = media_type

    @property
    def __raw__(self) -> Any:
        """Get the value from the ``MediaItem`` instance."""
        return self.media

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``MediaItem`` instance."""
        return {"representation": {"media_type": self.media_type, "value": self.media}}

    @classmethod
    def factory(
        cls,
        media: str,
        /,
        media_type: str = MediaType.MARKDOWN.value,
    ) -> MediaItem:
        """
        Create a new ``MediaItem``.

        Parameters
        ----------
        media : str
            The media to serialize.
        media_type : str
            The MIME type of the media content.

        Returns
        -------
        MediaItem
            A new ``MediaItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``media`` is not an instance of ``str``.
        ValueError
            If ``media_type`` is not valid.
        """
        if not isinstance(media, str):
            raise ItemTypeError(f"Type '{media.__class__}' is not supported.")

        # | Before Python 3.12, a TypeError is raised if a non-Enum-member is used in a
        # | containment check.
        #
        # https://docs.python.org/3.12/library/enum.html#enum.EnumType.__contains__
        if media_type not in MediaType._value2member_map_:
            raise ValueError(f"MIME type '{media_type}' is not supported.")

        return cls(media, media_type)



"""
PandasDataFrameItem.

This module defines the ``PandasDataFrameItem`` class used to serialize instances of
``pandas.DataFrame``, using the ``JSON`` format.
"""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Literal

from .item import Item, ItemTypeError, lazy_is_instance

if TYPE_CHECKING:
    import pandas


class PandasDataFrameItem(Item):
    """Serialize instances of ``pandas.DataFrame``, using the ``JSON`` format."""

    ORIENT: Literal["split"] = "split"

    def __init__(self, index_json_str: str, dataframe_json_str: str):
        """
        Initialize a ``PandasDataFrameItem``.

        Parameters
        ----------
        index_json_str : str
            The index of the ``pandas.DataFrame`` serialized in a str in the ``JSON``
            format.
        dataframe_json_str : str
            The ``pandas.DataFrame`` serialized in a str in the ``JSON`` format, without
            its index.
        """
        self.index_json_str = index_json_str
        self.dataframe_json_str = dataframe_json_str

    @cached_property
    def __raw__(self) -> pandas.DataFrame:
        """
        Get the value from the ``PandasDataFrameItem``.

        Notes
        -----
        Its content can slightly differ from the original because it has been serialized
        using ``pandas.to_json`` function and not pickled, in order to be
        environment-independent.
        """
        import io

        import pandas

        with (
            io.StringIO(self.index_json_str) as index_stream,
            io.StringIO(self.dataframe_json_str) as df_stream,
        ):
            index = pandas.read_json(index_stream, orient=self.ORIENT, dtype=False)
            index = index.set_index(list(index.columns))
            dataframe = pandas.read_json(df_stream, orient=self.ORIENT, dtype=False)
            dataframe.index = index.index

            return dataframe

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``PandasDataFrameItem`` instance."""
        return {
            "representation": {
                "media_type": "application/vnd.dataframe",
                "value": self.__raw__.fillna("NaN").to_dict(orient="tight"),
            }
        }

    @classmethod
    def factory(cls, value: pandas.DataFrame, /) -> PandasDataFrameItem:
        """
        Create a new ``PandasDataFrameItem`` from an instance of ``pandas.DataFrame``.

        It uses the ``JSON`` format.

        Parameters
        ----------
        value : ``pandas.DataFrame``
            The value to serialize.

        Returns
        -------
        PandasDataFrameItem
            A new ``PandasDataFrameItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` is not an instance of ``pandas.DataFrame``.

        Notes
        -----
        The dataframe must be JSON serializable.
        """
        if not lazy_is_instance(value, "pandas.core.frame.DataFrame"):
            raise ItemTypeError(f"Type '{value.__class__}' is not supported.")

        # Two native methods are available to serialize dataframe with multi-index,
        # while keeping the index names:
        #
        # 1. Using table orientation with JSON serializer:
        #    ```python
        #    json = dataframe.to_json(orient="table")
        #    dataframe = pandas.read_json(json, orient="table", dtype=False)
        #    ```
        #
        #    This method fails when an index/column name is an integer.
        #
        # 2. Using record orientation with indexes as columns:
        #    ```python
        #    dataframe = dataframe.reset_index()
        #    json = dataframe.to_json(orient="records")
        #    dataframe = pandas.read_json(json, orient="records", dtype=False)
        #    ```
        #
        #    This method fails when the index has the same name as one of the columns.
        #
        # None of those methods being compatible, we decide to store indexes separately.

        index = value.index.to_frame(index=False)
        dataframe_without_index = value.reset_index(drop=True)

        instance = cls(
            index.to_json(orient=PandasDataFrameItem.ORIENT),
            dataframe_without_index.to_json(orient=PandasDataFrameItem.ORIENT),
        )
        instance.__raw__ = value

        return instance



"""Login to ``skore hub``."""

from __future__ import annotations

import webbrowser
from datetime import datetime, timezone
from time import sleep

from httpx import HTTPError
from rich.align import Align
from rich.panel import Panel

from .. import console
from ..client import api
from ..client.client import AuthenticationError
from .otp_server import OTPServer
from .token import Token


def refresh():
    """Attempt to refresh an existing token.

    Returns
    -------
    tuple
        A tuple containing:
        - token: Token
            The token valid or not
        - is_valid: bool
            `True` if the (potentially) refreshed token is valid
            `False` otherwise.
    """
    token = Token()
    if not token.valid:
        return False, token

    now = datetime.now(timezone.utc)
    if token.expires_at <= now:
        # we have a token but it's expired
        # try to refresh it
        try:
            token.refresh()
            return True, token
        except HTTPError:
            return False, token

    # we have a token and it is not expired
    return True, token


def login(timeout=600, auto_otp=True, port=0):
    """Login to the skore-HUB."""
    is_valid, token = refresh()
    if is_valid:
        return token
    if auto_otp:
        access_token, refresh_token, expires_at = None, None, None

        def callback(state):
            nonlocal access_token
            nonlocal refresh_token
            nonlocal expires_at

            api.post_oauth_device_callback(state=state, user_code=user_code)
            access_token, refresh_token, expires_at = api.get_oauth_device_token(
                device_code=device_code
            )

        server = OTPServer(callback=callback)

        try:
            server.start(port=port)
            authorization_url, device_code, user_code = api.get_oauth_device_login(
                f"http://localhost:{server.port}"
            )

            webbrowser.open(authorization_url)

            start = datetime.now()

            while access_token is None or refresh_token is None or expires_at is None:
                if (datetime.now() - start).total_seconds() > timeout:
                    raise AuthenticationError("Timeout") from None

                sleep(0.25)
        finally:
            server.stop()

        return Token(
            access_token=access_token,
            refresh_token=refresh_token,
            expires_at=expires_at,
        )

    else:
        # Request a new authorization URL
        authorization_url, device_code, user_code = api.get_oauth_device_login()

        # Display authentication info to the user
        console.print(
            "\n"
            "🌍 Opening your default browser to start the authentication process.\n"
            "❔ If your browser did not open visit our "
            f"[link={authorization_url}]authentication page[/link].\n"
        )
        console.print(
            Panel(
                Align(f"[bold]{user_code}[/bold]", align="center"),
                title="[cyan]Your unique code is[/cyan]",
                border_style="orange1",
                expand=False,
                padding=1,
                title_align="center",
            )
        )
        # Open the default browser
        webbrowser.open(authorization_url)
        start = datetime.now()

        # Start polling Skore-Hub, waiting for the token
        while True:
            try:
                return Token(*api.get_oauth_device_token(device_code=device_code))
            except HTTPError:
                sleep(0.5)

                if (datetime.now() - start).total_seconds() > timeout:
                    raise AuthenticationError("Timeout") from None



"""Abstract base class for all items in the ``skore`` hub project."""

from __future__ import annotations

from abc import ABC, abstractmethod
from base64 import b64decode, b64encode
from contextlib import contextmanager
from inspect import signature as inspect_signature
from typing import Any


@contextmanager
def switch_mpl_backend():
    """
    Context-manager for switching ``matplotlib.backend`` to ``agg``.

    Notes
    -----
    The ``agg`` backend is a non-interactive backend that can only write to files.
    It is used in ``skore-hub-project`` to generate artifacts where we don't need an
    X display.

    https://matplotlib.org/stable/users/explain/figure/backends.html#selecting-a-backend
    """
    import matplotlib

    original_backend = matplotlib.get_backend()

    try:
        matplotlib.use("agg")
        yield
    finally:
        matplotlib.use(original_backend)


def lazy_is_instance(value: Any, cls_fullname: str) -> bool:
    """Return True if value is an instance of ``cls_fullname``."""
    return cls_fullname in {
        f"{cls.__module__}.{cls.__name__}" for cls in value.__class__.__mro__
    }


def bytes_to_b64_str(literal: bytes) -> str:
    """Encode the bytes-like object ``literal`` in a Base64 str."""
    return b64encode(literal).decode("utf-8")


def b64_str_to_bytes(literal: str) -> bytes:
    """Decode the Base64 str object ``literal`` in a bytes."""
    return b64decode(literal.encode("utf-8"))


class ItemTypeError(Exception):
    """
    Item type exception.

    Exception raised when an attempt is made to convert an object to an ``Item`` with an
    unsupported type.
    """


class Item(ABC):
    """
    Abstract base class for all items in the ``skore`` hub project.

    This class provides a common interface for all items, including the serialization of
    the parameters needed to recreate the instance from the hub project.

    ``Item`` is an internal concept that is used as a DTO (Data Transfer Object) to
    exchange python objects between ``skore`` and ``skore hub``.
    """

    @property
    def __parameters__(self) -> dict[str, dict[str, Any]]:
        """
        Get the parameters of the ``Item`` instance.

        These parameters must be sufficient to recreate the instance.
        They are persisted in the ``skore`` hub project and retrieved as needed.
        """
        cls = self.__class__
        cls_name = cls.__name__
        cls_parameters = inspect_signature(cls).parameters

        return {
            "parameters": {
                "class": cls_name,
                "parameters": {p: getattr(self, p) for p in cls_parameters},
            }
        }

    @property
    def __metadata__(self) -> dict[str, Any]:
        """Get the metadata of the ``Item`` instance."""
        return dict()

    @property
    @abstractmethod
    def __raw__(self) -> Any:
        """Get the raw python object from the ``Item`` instance."""

    @property
    @abstractmethod
    def __representation__(self) -> dict[str, Any]:
        """Get the representation of the ``Item`` instance."""

    @classmethod
    @abstractmethod
    def factory(cls, *args, **kwargs) -> Item:
        """Create and return a new instance of ``Item``."""



"""Class definition of the ``skore`` hub project."""

from __future__ import annotations

from functools import cached_property
from operator import itemgetter
from types import SimpleNamespace
from typing import TYPE_CHECKING

from .. import item as item_module
from ..client.client import AuthenticatedClient
from ..item.item import lazy_is_instance

if TYPE_CHECKING:
    from typing import TypedDict, Union

    from skore.sklearn import EstimatorReport

    class Metadata(TypedDict):  # noqa: D101
        id: str
        run_id: str
        key: str
        date: str
        learner: str
        dataset: str
        ml_task: str
        rmse: Union[float, None]
        log_loss: Union[float, None]
        roc_auc: Union[float, None]
        fit_time: float
        predict_time: float


class Project:
    """
    API to manage a collection of key-report pairs persisted in a hub storage.

    It communicates with the Probabl's ``skore hub`` storage, based on the pickle
    representation. Its constructor initializes a hub project by creating a new
    project or by loading an existing one from a defined tenant.

    The class main methods are :func:`~skore_hub_project.Project.put`,
    :func:`~skore_hub_project.reports.metadata` and
    :func:`~skore_hub_project.reports.get`, respectively to insert a key-report pair
    into the Project, to obtain the reports metadata and to get a specific report.

    Parameters
    ----------
    tenant : str
        The tenant of the project.

        A tenant is a ``skore hub`` concept that must be configured on the
        ``skore hub`` interface. It represents an isolated entity managing users,
        projects, and resources. It can be a company, organization, or team that
        operates independently within the system.
    name : str
        The name of the project.

    Attributes
    ----------
    tenant : str
        The tenant of the project.
    name : str
        The name of the project.
    run_id : str
        The current run identifier of the project.
    """

    def __init__(self, tenant: str, name: str):
        """
        Initialize a hub project.

        Initialize a hub project by creating a new project or by loading an existing
        one from a defined tenant.

        Parameters
        ----------
        tenant : Path
            The tenant of the project.

            A tenant is a ``skore hub`` concept that must be configured on the
            ``skore hub`` interface. It represents an isolated entity managing users,
            projects, and resources. It can be a company, organization, or team that
            operates independently within the system.
        name : str
            The name of the project.
        """
        self.__tenant = tenant
        self.__name = name

    @property
    def tenant(self) -> str:
        """The tenant of the project."""
        return self.__tenant

    @property
    def name(self) -> str:
        """The name of the project."""
        return self.__name

    @cached_property
    def run_id(self) -> str:
        """The current run identifier of the project."""
        with AuthenticatedClient(raises=True) as client:
            request = client.post(f"projects/{self.tenant}/{self.name}/runs")
            run = request.json()

            return run["id"]

    def put(self, key: str, report: EstimatorReport):
        """
        Put a key-report pair to the hub project.

        If the key already exists, its last report is modified to point to this new
        report, while keeping track of the report history.

        Parameters
        ----------
        key : str
            The key to associate with ``report`` in the hub project.
        report : skore.EstimatorReport
            The report to associate with ``key`` in the hub project.

        Raises
        ------
        TypeError
            If the combination of parameters are not valid.
        """
        if not isinstance(key, str):
            raise TypeError(f"Key must be a string (found '{type(key)}')")

        if not lazy_is_instance(
            report, "skore.sklearn._estimator.report.EstimatorReport"
        ):
            raise TypeError(
                f"Report must be a `skore.EstimatorReport` (found '{type(report)}')"
            )

        item = item_module.object_to_item(report)

        with AuthenticatedClient(raises=True) as client:
            client.post(
                f"projects/{self.tenant}/{self.name}/items",
                json={
                    **item.__metadata__,
                    **item.__representation__,
                    **item.__parameters__,
                    "key": key,
                    "run_id": self.run_id,
                },
            )

    @property
    def reports(self):
        """Accessor for interaction with the persisted reports."""

        def get(id: str) -> EstimatorReport:
            """Get a persisted report by its id."""

            def dto(report):
                item_class_name = report["raw"]["class"]
                item_class = getattr(item_module, item_class_name)
                item_parameters = report["raw"]["parameters"]
                item = item_class(**item_parameters)
                return item.__raw__

            with AuthenticatedClient(raises=True) as client:
                response = client.get(
                    f"projects/{self.tenant}/{self.name}/experiments/estimator-reports/{id}"
                )

            return dto(response.json())

        def metadata() -> list[Metadata]:
            """Obtain metadata for all persisted reports regardless of their run."""

            def dto(summary):
                metrics = {
                    metric["name"]: metric["value"]
                    for metric in summary["metrics"]
                    if metric["data_source"] in (None, "test")
                }

                return {
                    "id": summary["id"],
                    "run_id": summary["run_id"],
                    "key": summary["key"],
                    "date": summary["created_at"],
                    "learner": summary["estimator_class_name"],
                    "dataset": summary["dataset_fingerprint"],
                    "ml_task": summary["ml_task"],
                    "rmse": metrics.get("rmse"),
                    "log_loss": metrics.get("log_loss"),
                    "roc_auc": metrics.get("roc_auc"),
                    "fit_time": metrics.get("fit_time"),
                    "predict_time": metrics.get("predict_time"),
                }

            with AuthenticatedClient(raises=True) as client:
                response = client.get(
                    f"projects/{self.tenant}/{self.name}/experiments/estimator-reports"
                )

            return sorted(map(dto, response.json()), key=itemgetter("date"))

        # Ensure project is created by calling `self.run_id`
        return self.run_id and SimpleNamespace(get=get, metadata=metadata)

    def __repr__(self) -> str:  # noqa: D105
        return f"Project(hub://{self.tenant}@{self.name})"



"""
PickleItem.

This module defines the ``PickleItem`` class used to serialize objects that cannot be
otherwise, using binary protocols.
"""

from __future__ import annotations

from functools import cached_property
from io import BytesIO
from typing import Any, TypeVar

from joblib import dump, load

from .item import Item, b64_str_to_bytes, bytes_to_b64_str

# Create a generic variable that can be `PickleItem`, or any subclass.
T = TypeVar("T", bound="PickleItem")


class PickleItem(Item):
    """Serialize objects that cannot be otherwise, using binary protocols."""

    def __init__(self, pickle_b64_str: str):
        """
        Initialize a ``PickleItem``.

        Parameters
        ----------
        pickle_b64_str : str
            The raw bytes of the pickled object encoded in a base64 str.
        """
        self.pickle_b64_str = pickle_b64_str

    @cached_property
    def __raw__(self) -> Any:
        """Get the deserialized python object from the ``PickleItem`` instance."""
        pickle_bytes = b64_str_to_bytes(self.pickle_b64_str)

        with BytesIO(pickle_bytes) as stream:
            return load(stream)

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``PickleItem`` instance."""
        return {
            "representation": {
                "media_type": "text/markdown",
                "value": f"```python\n{repr(self.__raw__)}\n```",
            }
        }

    @classmethod
    def factory(cls: type[T], value: Any, /) -> T:
        """
        Create a new ``PickleItem`` from ``value`` using binary protocols.

        Parameters
        ----------
        value: Any
            The value to serialize.

        Returns
        -------
        PickleItem
            A new ``PickleItem`` instance.
        """
        with BytesIO() as stream:
            dump(value, stream)

            pickle_bytes = stream.getvalue()
            pickle_b64_str = bytes_to_b64_str(pickle_bytes)

        instance = cls(pickle_b64_str)
        instance.__raw__ = value

        return instance



"""Client exchanging with ``skore hub``."""

from __future__ import annotations

from contextlib import suppress
from datetime import datetime, timezone
from functools import cached_property
from urllib.parse import urljoin

from httpx import URL, Client, HTTPStatusError, Response

from ..authentication.token import Token
from .api import URI


class AuthenticationError(Exception):
    """An exception dedicated to authentication."""


class AuthenticatedClient(Client):
    """Client exchanging with ``skore hub``."""

    ERROR_TYPES = {
        1: "Informational response",
        3: "Redirect response",
        4: "Client error",
        5: "Server error",
    }

    def __init__(self, *, raises=False):
        super().__init__(follow_redirects=True, timeout=300)

        self.raises = raises

    @cached_property
    def token(self):
        """Access token."""
        return Token()

    def request(self, method: str, url: URL | str, **kwargs) -> Response:
        """Execute request with access token, and refresh the token if needed."""
        # Check if token is well initialized
        if not self.token.valid:
            raise AuthenticationError(
                "You are not logged in. Please run `skore-hub-login`"
            )

        # Check if token must be refreshed
        if self.token.expires_at <= datetime.now(timezone.utc):
            self.token.refresh()

        # Overload headers with authorization token
        headers = kwargs.pop("headers", None) or {}
        headers["Authorization"] = f"Bearer {self.token.access_token}"
        response = super().request(
            method,
            urljoin(URI, str(url)),
            headers=headers,
            **kwargs,
        )

        if self.raises and not response.is_success:
            status_class = response.status_code // 100
            error_type = self.ERROR_TYPES.get(status_class, "Invalid status code")
            message = (
                f"{error_type} '{response.status_code} {response.reason_phrase}' "
                f"for url '{response.url}'"
            )

            with suppress(Exception):
                message += f": {response.json()['message']}"

            raise HTTPStatusError(message, request=response.request, response=response)

        return response



"""
SkoreEstimatorReportItem.

This module defines the ``SkoreEstimatorReportItem`` class used to serialize instances
of ``skore.EstimatorReport``, using binary protocols.
"""

from __future__ import annotations

from contextlib import suppress
from inspect import getmembers, ismethod, signature
from math import isfinite
from operator import attrgetter
from typing import TYPE_CHECKING

from .item import ItemTypeError, lazy_is_instance, switch_mpl_backend
from .matplotlib_figure_item import MatplotlibFigureItem
from .media_item import MediaItem
from .pandas_dataframe_item import PandasDataFrameItem
from .pickle_item import PickleItem

if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Any, Literal, TypedDict, Union

    from skore.sklearn import EstimatorReport

    class MetadataFunction:  # noqa: D101
        metadata: Any

    class Metric(TypedDict):  # noqa: D101
        name: str
        verbose_name: Union[str, None]
        value: float
        data_source: Union[str, None]
        greater_is_better: Union[bool, None]
        position: Union[int, None]

    class Representation(TypedDict):  # noqa: D101
        key: str
        verbose_name: Union[str, None]
        category: Literal["performance", "feature_importance", "model"]
        attributes: dict
        representation: dict
        parameters: dict


def cast_to_float(value: Any) -> Union[float, None]:
    """Cast value to float."""
    with suppress(TypeError):
        if (value := float(value)) and isfinite(value):
            return value

    return None


def metadata_function(function: Any) -> MetadataFunction:
    """
    Decorate function that has to be annotated as ``metadata``.

    Annotated functions as ``medatata`` are dynamically retrieved and called at runtime
    to compute a snapshot of all the available metadata of a report.

    Notes
    -----
    This decorator uses the `function attributes <https://peps.python.org/pep-0232/>`_.
    """
    function.metadata = ...
    return function


class Metadata:
    """
    Compute report's metadata to exchange with ``skore hub``.

    Notes
    -----
    To add a new metadata to compute, you have to add a new method decorated as
    ``metadata``. Its name is used as identifier.
    """

    def __init__(self, report: EstimatorReport):
        """
        Initialize a ``Metadata``.

        Parameters
        ----------
        report : ``skore.EstimatorReport``
            The report from which computing the metadata.
        """
        self.report = report

    @metadata_function
    def estimator_class_name(self) -> str:
        """Return the name of the report's estimator."""
        return self.report.estimator_name_

    @metadata_function
    def estimator_hyper_params(self) -> dict:
        """DeprecationWarning: send empty dictionary to not break the hub API."""
        return {}

    @metadata_function
    def dataset_fingerprint(self) -> str:
        """Return the hash of the targets in the test-set."""
        import joblib

        return joblib.hash(self.report.y_test)

    @metadata_function
    def ml_task(self) -> str:
        """Return the type of ML task covered by the report."""
        return self.report._ml_task

    @metadata_function
    def metrics(self) -> list[Metric]:
        """
        Return the list of scalar metrics that can be computed from the report.

        Notes
        -----
        Unavailable metrics are automatically filtered out.

        All metrics whose value is not a scalar are currently ignored:
        - ignore ``list[float]`` for multi-output ML task,
        - ignore ``dict[str: float]`` for multi-classes ML task.

        The position field is used to drive the HUB's parallel coordinates plot:
        - int [0, inf[, to be displayed at the position,
        - None, not to be displayed.
        """

        def metric(
            name: str,
            verbose_name: str,
            data_source: str,
            greater_is_better: bool,
            position: Union[int, None],
            /,
        ) -> Union[Metric, None]:
            if hasattr(self.report.metrics, name):
                value = getattr(self.report.metrics, name)(data_source=data_source)

                if (value := cast_to_float(value)) is not None:
                    return {
                        "name": name,
                        "verbose_name": verbose_name,
                        "value": value,
                        "data_source": data_source,
                        "greater_is_better": greater_is_better,
                        "position": position,
                    }

            return None

        def timing(
            name: str,
            verbose_name: str,
            data_source: Union[str, None],
            greater_is_better: bool,
            position: Union[int, None],
            /,
        ) -> Union[Metric, None]:
            timings = self.report.metrics.timings()
            value = timings.get(
                name if name != "predict_time" else f"{name}_{data_source}"
            )

            if (value := cast_to_float(value)) is not None:
                return {
                    "name": name,
                    "verbose_name": verbose_name,
                    "value": value,
                    "data_source": data_source,
                    "greater_is_better": greater_is_better,
                    "position": position,
                }

            return None

        return list(
            filter(
                None,
                (
                    metric("accuracy", "Accuracy", "train", True, None),
                    metric("accuracy", "Accuracy", "test", True, None),
                    metric("brier_score", "Brier score", "train", False, None),
                    metric("brier_score", "Brier score", "test", False, None),
                    metric("log_loss", "Log loss", "train", False, 4),
                    metric("log_loss", "Log loss", "test", False, 4),
                    metric("precision", "Precision", "train", True, None),
                    metric("precision", "Precision", "test", True, None),
                    metric("r2", "R²", "train", True, None),
                    metric("r2", "R²", "test", True, None),
                    metric("recall", "Recall", "train", True, None),
                    metric("recall", "Recall", "test", True, None),
                    metric("rmse", "RMSE", "train", False, 3),
                    metric("rmse", "RMSE", "test", False, 3),
                    metric("roc_auc", "ROC AUC", "train", True, 3),
                    metric("roc_auc", "ROC AUC", "test", True, 3),
                    # timings must be calculated last
                    timing("fit_time", "Fit time (s)", None, False, 1),
                    timing("predict_time", "Predict time (s)", "train", False, 2),
                    timing("predict_time", "Predict time (s)", "test", False, 2),
                ),
            )
        )

    def __iter__(self) -> Generator[tuple[str, Any]]:
        """
        Dynamically generate metadata by calling all ``metadata`` methods.

        Notes
        -----
        Null metadata are automatically filtered out.
        """
        for key, method in getmembers(self):
            if (
                ismethod(method)
                and hasattr(method, "metadata")
                and ((value := method()) is not None)
            ):
                yield (key, value)


class Representations:
    """Compute report's representation to exchange with ``skore hub``."""

    def __init__(self, report):
        """
        Initialize a ``Metadata``.

        Parameters
        ----------
        report : ``skore.EstimatorReport``
            The report from which computing the metadata.
        """
        self.report = report

    def mpl(
        self,
        name,
        verbose_name,
        category,
        **kwargs,
    ) -> Union[Representation, None]:
        """Return sub-representation made of ``matplotlib`` figures."""
        try:
            function = attrgetter(name)(self.report)
        except AttributeError:
            return None
        else:
            function_parameters = signature(function).parameters
            function_kwargs = {
                k: v for k, v in kwargs.items() if k in function_parameters
            }

            display = function(**function_kwargs)

            with switch_mpl_backend():
                display.plot()

            item = MatplotlibFigureItem.factory(display.figure_)

            return {
                "key": name.split(".")[-1],
                "verbose_name": verbose_name,
                "category": category,
                "attributes": kwargs,
                "parameters": {},
                "representation": item.__representation__["representation"],
            }

    def pd(self, name, verbose_name, category, **kwargs) -> Union[Representation, None]:
        """Return sub-representation made of ``pandas`` dataframes."""
        try:
            function = attrgetter(name)(self.report)
        except AttributeError:
            return None
        else:
            function_parameters = signature(function).parameters
            function_kwargs = {
                k: v for k, v in kwargs.items() if k in function_parameters
            }
            dataframe = function(**function_kwargs)
            item = PandasDataFrameItem.factory(dataframe)

            return {
                "key": name.split(".")[-1],
                "verbose_name": verbose_name,
                "category": category,
                "attributes": kwargs,
                "parameters": {},
                "representation": item.__representation__["representation"],
            }

    def estimator_html_repr(self) -> Representation:
        """Return ``sklearn`` HTML representation of the report's estimator."""
        import sklearn.utils

        item = MediaItem.factory(
            sklearn.utils.estimator_html_repr(self.report.estimator_),
            media_type="text/html",
        )

        return {
            "key": "estimator_html_repr",
            "verbose_name": None,
            "category": "model",
            "attributes": {},
            "parameters": {},
            "representation": item.__representation__["representation"],
        }

    def __iter__(self) -> Generator[Representation]:
        """
        Dynamically generate representation, with report's parameters combinations.

        Notes
        -----
        Null representations are automatically filtered out.
        """
        yield from filter(
            None,
            (
                self.mpl(
                    "metrics.precision_recall",
                    "Precision Recall",
                    "performance",
                    data_source="train",
                ),
                self.mpl(
                    "metrics.precision_recall",
                    "Precision Recall",
                    "performance",
                    data_source="test",
                ),
                self.mpl(
                    "metrics.prediction_error",
                    "Prediction error",
                    "performance",
                    data_source="train",
                ),
                self.mpl(
                    "metrics.prediction_error",
                    "Prediction error",
                    "performance",
                    data_source="test",
                ),
                self.mpl("metrics.roc", "ROC", "performance", data_source="train"),
                self.mpl("metrics.roc", "ROC", "performance", data_source="test"),
                self.pd(
                    "feature_importance.permutation",
                    "Feature importance - Permutation",
                    "feature_importance",
                    data_source="train",
                    method="permutation",
                ),
                self.pd(
                    "feature_importance.permutation",
                    "Feature importance - Permutation",
                    "feature_importance",
                    data_source="test",
                    method="permutation",
                ),
                self.pd(
                    "feature_importance.mean_decrease_impurity",
                    "Feature importance - Mean Decrease Impurity (MDI)",
                    "feature_importance",
                    method="mean_decrease_impurity",
                ),
                self.pd(
                    "feature_importance.coefficients",
                    "Feature importance - Coefficients",
                    "feature_importance",
                    method="coefficients",
                ),
                self.estimator_html_repr(),
            ),
        )


class SkoreEstimatorReportItem(PickleItem):
    """Serialize instances of ``skore.EstimatorReport``, using binary protocols."""

    @property
    def __metadata__(self) -> dict[str, Any]:
        """Get the metadata of the ``SkoreEstimatorReportItem`` instance."""
        return dict(Metadata(self.__raw__))

    @property
    def __representation__(self) -> dict[str, list[Representation]]:
        """Get the representation of the ``SkoreEstimatorReportItem`` instance."""
        return {"related_items": list(Representations(self.__raw__))}

    @classmethod
    def factory(cls, value: EstimatorReport, /) -> SkoreEstimatorReportItem:
        """
        Create a new ``SkoreEstimatorReportItem``.

        Create a new ``SkoreEstimatorReportItem`` from an instance of
        ``skore.EstimatorReport``.

        Parameters
        ----------
        value : ``skore.EstimatorReport``
            The value to serialize.

        Returns
        -------
        SkoreEstimatorReportItem
            A new ``SkoreEstimatorReportItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` is not an instance of ``skore.EstimatorReport``.
        """
        if lazy_is_instance(value, "skore.sklearn._estimator.report.EstimatorReport"):
            return super().factory(value)

        raise ItemTypeError(f"Type '{value.__class__}' is not supported.")



"""
PlotlyFigureItem.

This module defines the ``PlotlyFigureItem`` class used to serialize instances of
``plotly`` figures, using the ``JSON`` format.
"""

from __future__ import annotations

from functools import cached_property
from json import loads
from typing import TYPE_CHECKING

from .item import Item, ItemTypeError, lazy_is_instance

if TYPE_CHECKING:
    import plotly.basedatatypes


class PlotlyFigureItem(Item):
    """Serialize instances of ``plotly`` figures, using the ``JSON`` format."""

    def __init__(self, figure_json_str: str):
        """
        Initialize a ``PlotlyFigureItem``.

        Parameters
        ----------
        figure_json_str : str
            The ``plotly`` figure serialized in a str in the ``JSON`` format.
        """
        self.figure_json_str = figure_json_str

    @cached_property
    def __raw__(self) -> plotly.basedatatypes.BaseFigure:
        """Get the value from the ``PlotlyFigureItem`` instance."""
        import plotly.io

        return plotly.io.from_json(self.figure_json_str)

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``PlotlyFigureItem`` instance."""
        return {
            "representation": {
                "media_type": "application/vnd.plotly.v1+json",
                "value": loads(self.figure_json_str),
            }
        }

    @classmethod
    def factory(cls, value: plotly.basedatatypes.BaseFigure, /) -> PlotlyFigureItem:
        """
        Create a new ``PlotlyFigureItem`` from an instance of ``plotly`` figure.

        It uses the ``JSON`` format.

        Parameters
        ----------
        value: ``plotly`` figure.
            The value to serialize.

        Returns
        -------
        PlotlyFigureItem
            A new ``PlotlyFigureItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` is not an instance of ``plotly`` figure.
        """
        if not lazy_is_instance(value, "plotly.basedatatypes.BaseFigure"):
            raise ItemTypeError(f"Type '{value.__class__}' is not supported.")

        import plotly.io

        instance = cls(plotly.io.to_json(value, engine="json"))
        instance.__raw__ = value

        return instance



"""Package that provides APIs to communicate between ``skore`` and ``skore hub``."""

from rich.console import Console
from rich.theme import Theme

from .project.project import Project

__all__ = ["Project", "console"]


console = Console(
    width=88,
    theme=Theme(
        {
            "repr.str": "cyan",
            "rule.line": "orange1",
            "repr.url": "orange1",
        }
    ),
)



"""
Module definition of the ``Item`` classes.

``Item`` is an internal concept that is used as a DTO (Data Transfer Object) to exchange
python objects between ``skore`` and ``skore hub``.
"""

from __future__ import annotations

from contextlib import suppress
from typing import Any

from .altair_chart_item import AltairChartItem
from .item import Item, ItemTypeError
from .jsonable_item import JSONableItem
from .matplotlib_figure_item import MatplotlibFigureItem
from .media_item import MediaItem, MediaType
from .numpy_array_item import NumpyArrayItem
from .pandas_dataframe_item import PandasDataFrameItem
from .pandas_series_item import PandasSeriesItem
from .pickle_item import PickleItem
from .pillow_image_item import PillowImageItem
from .plotly_figure_item import PlotlyFigureItem
from .polars_dataframe_item import PolarsDataFrameItem
from .polars_series_item import PolarsSeriesItem
from .sklearn_base_estimator_item import SklearnBaseEstimatorItem
from .skore_estimator_report_item import SkoreEstimatorReportItem
from .skrub_table_report_item import SkrubTableReportItem


def object_to_item(object: Any, /) -> Item:
    """Serialize any python object into an ``Item``."""
    for cls in (
        # Lexicographically sorted, the order of execution doesn't matter
        AltairChartItem,
        MatplotlibFigureItem,
        NumpyArrayItem,
        PandasDataFrameItem,
        PandasSeriesItem,
        PillowImageItem,
        PlotlyFigureItem,
        PolarsDataFrameItem,
        PolarsSeriesItem,
        SklearnBaseEstimatorItem,
        SkoreEstimatorReportItem,
        SkrubTableReportItem,
        # JSONable must be the penultimate
        JSONableItem,
    ):
        with suppress(ImportError, ItemTypeError):
            # ImportError:
            #     The factories are responsible to import third-party libraries in a
            #     lazy way. If library is missing, an ImportError exception will
            #     automatically be thrown.
            # ItemTypeError:
            #     The factories are responsible for checking that parameters are of
            #     the correct type. If not, they throw a ItemTypeError exception.
            return cls.factory(object)
    return PickleItem.factory(object)


__all__ = [
    "AltairChartItem",
    "Item",
    "JSONableItem",
    "MatplotlibFigureItem",
    "MediaItem",
    "MediaType",
    "NumpyArrayItem",
    "PandasDataFrameItem",
    "PandasSeriesItem",
    "PickleItem",
    "PillowImageItem",
    "PlotlyFigureItem",
    "PolarsDataFrameItem",
    "PolarsSeriesItem",
    "SklearnBaseEstimatorItem",
    "SkoreEstimatorReportItem",
    "SkrubTableReportItem",
    "object_to_item",
]



"""
PillowImageItem.

This module defines the ``PillowImageItem`` class used to serialize instances of
``pillow`` images, using binary protocols.
"""

from __future__ import annotations

from functools import cached_property
from io import BytesIO
from typing import TYPE_CHECKING

from .item import (
    Item,
    ItemTypeError,
    b64_str_to_bytes,
    bytes_to_b64_str,
    lazy_is_instance,
)

if TYPE_CHECKING:
    import PIL.Image


class PillowImageItem(Item):
    """Serialize instances of ``pillow`` images, using binary protocols."""

    def __init__(
        self,
        image_b64_str: str,
        image_mode: str,
        image_size: tuple[int, int],
    ):
        """
        Initialize a ``PillowImageItem``.

        Parameters
        ----------
        image_b64_str : str
            The raw bytes of the image in the ``pillow`` serialization format, encoded
            in a base64 string.
        image_mode : str
            The image mode.
        image_size : tuple of int
            The image size.
        """
        self.image_b64_str = image_b64_str
        self.image_mode = image_mode
        self.image_size = image_size

    @cached_property
    def __raw__(self) -> PIL.Image.Image:
        """Get the value from the ``PillowImageItem`` instance."""
        import PIL.Image

        image_bytes = b64_str_to_bytes(self.image_b64_str)

        return PIL.Image.frombytes(
            mode=self.image_mode,
            size=self.image_size,
            data=image_bytes,
        )

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``PillowImageItem`` instance."""
        with BytesIO() as stream:
            self.__raw__.save(stream, format="png")

            png_bytes = stream.getvalue()
            png_b64_str = bytes_to_b64_str(png_bytes)

        return {
            "representation": {
                "media_type": "image/png;base64",
                "value": png_b64_str,
            }
        }

    @classmethod
    def factory(cls, value: PIL.Image.Image, /) -> PillowImageItem:
        """
        Create a new ``PillowImageItem`` from an instance of ``pillow`` image.

        It uses binary protocols.

        Parameters
        ----------
        value: ``pillow`` image.
            The value to serialize.

        Returns
        -------
        PillowImageItem
            A new ``NumpyArrayItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` is not an instance of ``pillow`` image.
        """
        if not lazy_is_instance(value, "PIL.Image.Image"):
            raise ItemTypeError(f"Type '{value.__class__}' is not supported.")

        image_bytes = value.tobytes()
        image_b64_str = bytes_to_b64_str(image_bytes)

        instance = cls(image_b64_str, value.mode, value.size)
        instance.__raw__ = value

        return instance



"""Token used for ``skore hub`` authentication."""

from __future__ import annotations

import functools
import json
import pathlib
import tempfile
from datetime import datetime
from typing import Optional

from ..client.api import post_oauth_refresh_token


class Token:
    """Wrap access, refresh and expires_at."""

    def __init__(
        self,
        access_token: Optional[str] = None,
        refresh_token: Optional[str] = None,
        expires_at: Optional[str] = None,
    ):
        if (
            access_token is not None
            and refresh_token is not None
            and expires_at is not None
        ):
            self.access_token = access_token
            self.refresh_token = refresh_token
            self.expires_at = expires_at

            # Save the tokens to the disk to prevent user to login more than once, as
            # long as the token is valid or can be refreshed.
            self.filepath.write_text(
                json.dumps(
                    (
                        access_token,
                        refresh_token,
                        expires_at,
                    )
                )
            )
        else:
            try:
                self.access_token, self.refresh_token, self.expires_at = json.loads(
                    self.filepath.read_text()
                )
            except FileNotFoundError:
                self.valid = False
                return

        # Flag used to ensure the validity of the token when requesting an URL with
        # httpx, in a lazy way.
        self.valid = True

    @property
    def expires_at(self):
        """Datetime of expiration of the tokens."""
        return self.__expires_at

    @expires_at.setter
    def expires_at(self, value: str):
        self.__expires_at = datetime.fromisoformat(value)

    @functools.cached_property
    def filepath(self):
        """Filepath used to save the tokens on disk."""
        return pathlib.Path(tempfile.gettempdir(), "skore.token")

    def refresh(self):
        """Call the API to get a fresh access token."""
        self.__init__(*post_oauth_refresh_token(self.refresh_token))

    def __repr__(self):
        """Repr implementation."""
        return f"Token('{self.access_token:.10}[...]')"



"""
AltairChartItem.

This module defines the ``AltairChartItem`` class used to serialize instances of
``altair`` charts, using the ``JSON`` format.
"""

from __future__ import annotations

from json import loads
from typing import TYPE_CHECKING

from .item import Item, ItemTypeError, lazy_is_instance

if TYPE_CHECKING:
    from altair.vegalite.v5.schema.core import TopLevelSpec as AltairChart


class AltairChartItem(Item):
    """Serialize instances of ``altair`` charts, using the ``JSON`` format."""

    def __init__(self, chart_json_str: str):
        """
        Initialize a ``AltairChartItem``.

        Parameters
        ----------
        chart_json_str : str
            The ``altair`` chart serialized in a str in the ``JSON`` format.
        """
        self.chart_json_str = chart_json_str

    @property
    def __raw__(self) -> AltairChart:
        """Get the value from the ``AltairChartItem`` instance."""
        import altair

        return altair.Chart.from_json(self.chart_json_str)

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``AltairChartItem`` instance."""
        return {
            "representation": {
                "media_type": "application/vnd.vega.v5+json",
                "value": loads(self.chart_json_str),
            }
        }

    @classmethod
    def factory(cls, value: AltairChart, /) -> AltairChartItem:
        """
        Create a new ``AltairChartItem`` from an instance of ``altair`` chart.

        It uses the ``JSON`` format.

        Parameters
        ----------
        value: ``altair`` chart.
            The value to serialize.

        Returns
        -------
        AltairChartItem
            A new ``AltairChartItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` is not an instance of ``altair`` chart.
        """
        if not lazy_is_instance(value, "altair.vegalite.v5.schema.core.TopLevelSpec"):
            raise ItemTypeError(f"Type '{value.__class__}' is not supported.")

        return cls(value.to_json())



"""
PolarsDataFrameItem.

This module defines the ``PolarsDataFrameItem`` class used to serialize instances of
``polars.DataFrame``, using the ``JSON`` format.
"""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from .item import Item, ItemTypeError, lazy_is_instance

if TYPE_CHECKING:
    import polars


class PolarsDataFrameItem(Item):
    """Serialize instances of ``polars.DataFrame``, using the ``JSON`` format."""

    def __init__(self, dataframe_json_str: str):
        """
        Initialize a ``PolarsDataFrameItem``.

        Parameters
        ----------
        dataframe_json_str : str
            The ``polars.DataFrame`` serialized in a str in the ``JSON`` format.
        """
        self.dataframe_json_str = dataframe_json_str

    @cached_property
    def __raw__(self) -> polars.DataFrame:
        """
        Get the value from the ``PolarsDataFrameItem``.

        Notes
        -----
        Its content can slightly differ from the original because it has been serialized
        using ``polars.to_json`` function and not pickled, in order to be
        environment-independent.
        """
        import io

        import polars

        with io.StringIO(self.dataframe_json_str) as df_stream:
            return polars.read_json(df_stream)

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``PolarsDataFrameItem`` instance."""
        return {
            "representation": {
                "media_type": "application/vnd.dataframe",
                "value": self.__raw__.to_pandas().fillna("NaN").to_dict(orient="tight"),
            }
        }

    @classmethod
    def factory(cls, value: polars.DataFrame, /) -> PolarsDataFrameItem:
        """
        Create a new ``PolarsDataFrameItem`` from an instance of ``polars.DataFrame``.

        It uses the ``JSON`` format.

        Parameters
        ----------
        value : ``polars.DataFrame``
            The value to serialize.

        Returns
        -------
        PolarsDataFrameItem
            A new ``PolarsDataFrameItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` is not an instance of ``polars.DataFrame``.

        Notes
        -----
        The dataframe must be JSON serializable.
        """
        if not lazy_is_instance(value, "polars.dataframe.frame.DataFrame"):
            raise ItemTypeError(f"Type '{value.__class__}' is not supported.")

        instance = cls(value.write_json())
        instance.__raw__ = value

        return instance



"""
SkrubTableReportItem.

This module defines the ``SkrubTableReportItem`` class used to serialize instances of
``skrub.TableReport``, using binary protocols.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .item import ItemTypeError, lazy_is_instance, switch_mpl_backend
from .pickle_item import PickleItem

if TYPE_CHECKING:
    from skrub import TableReport


class SkrubTableReportItem(PickleItem):
    """Serialize instances of ``skrub.TableReport``, using binary protocols."""

    @property
    def __representation__(self) -> dict:
        """Get the representation of the ``SkrubTableReportItem`` instance."""
        with switch_mpl_backend():
            return {
                "representation": {
                    "media_type": "text/html",
                    "value": self.__raw__.html_snippet(),
                }
            }

    @classmethod
    def factory(cls, value: TableReport, /) -> SkrubTableReportItem:
        """
        Create a new ``SkrubTableReportItem`` from an instance of ``skrub.TableReport``.

        It uses binary protocols.

        Parameters
        ----------
        value : ``skrub.TableReport``
            The value to serialize.

        Returns
        -------
        SkrubTableReportItem
            A new ``SkrubTableReportItem`` instance.

        Raises
        ------
        ItemTypeError
            If ``value`` is not an instance of ``skrub.TableReport``.
        """
        if not lazy_is_instance(value, "skrub._reporting._table_report.TableReport"):
            raise ItemTypeError(f"Type '{value.__class__}' is not supported.")

        return super().factory(value)



"""Create an HTTP server which will wait for OTP success callback."""

from http import HTTPStatus
from http.server import BaseHTTPRequestHandler
from socketserver import TCPServer
from threading import Thread
from typing import Callable
from urllib.parse import parse_qs, urlparse

SUCCESS_PAGE = b"<p>You can now close this page.</p>"


def _get_handler(callback):
    class _Handler(BaseHTTPRequestHandler):
        def log_message(self, *args):
            """Disable logging."""

        def do_GET(self):
            parsed = urlparse(self.path)

            if parsed.path != "/":
                self.send_response(HTTPStatus.NO_CONTENT)
                self.send_header("Content-Length", 0)
                self.end_headers()
                self.wfile.write(b"")
            else:
                (state,) = parse_qs(parsed.query).get("state", [None])
                callback(state)

                self.send_response(HTTPStatus.OK)
                self.send_header("Content-type", "text/html")
                self.send_header("Content-Length", len(SUCCESS_PAGE))
                self.end_headers()
                self.wfile.write(SUCCESS_PAGE)

    return _Handler


class OTPServer:
    """HTTP server to handle OTP success callbacks.

    Parameters
    ----------
    callback : callable
        Function to be called when the server receives a request.
        The callback function should accept one parameter for the state.

    Attributes
    ----------
    port
        The port number the server is listening on.
    started
        The server is started, or not.
    """

    def __init__(self, callback: Callable[[str], None]):
        self.callback = callback

    def start(self, port=0):
        """Start the server.

        Parameters
        ----------
        port : int, default 0
            The port number the server is listening on.
            By default, we let the system choose a free port.
        """
        if not self.started:
            self.__tcp = TCPServer(("", port), _get_handler(self.callback))
            self.__thread = Thread(target=self.__tcp.serve_forever)
            self.__thread.start()

        return self

    @property
    def port(self):
        """Property indicating the port number the server is listening on."""
        try:
            return self.__tcp.server_address[1]
        except AttributeError:
            raise SystemError("Server not started") from None

    @property
    def started(self):
        """Property indicating whether the server is started or not."""
        return hasattr(self, "_OTPServer__tcp")

    def stop(self):
        """Stop the server."""
        if self.started:
            # Stop server and thread
            self.__tcp.shutdown()
            self.__thread.join()

            # Delete start related attributes
            del self.__tcp
            del self.__thread
